package com.toplink.test;

import java.util.List;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.EntityTransaction;
import javax.persistence.Persistence;
import javax.persistence.Query;

/**
 * Test harness for showing that TopLink loads dependent objects multiple
 * times when building an object following a query.  In the example
 * below a list of MailMessage entities is returned.  Each MailMessage
 * has a ManyToOne relationship with the User entity who created it.
 * 
 * In the example below two mail messages are created for user1.
 * 
 * When we request messages for user1, I would expect to see SELECT statements to load:
 * 1. Messages for user1
 * 2. User 1 (This is to set the embedded User referenced in each message to support the
 *    ManyToOne relationship (MailMessage->User).  This is probably unnecessary
 *    as well given that we've already loaded user1 in this UnitOfWork.
 *
 * Instead I see SELECT statements to load:
 * 1. Messages for user1
*  2. User 1
*  3. User 1 (definitely unnecessary)
 * 
 * To execute the test:
 * 
 * 1. Run the example first with any argument on the command line
 * to force it to create the initial user and mail data.  
 * 
 * 2. Comment out the toplink.ddl-generation property from persistence.xml so
 * it does not re-create the tables, then run it without any
 * command line arguments to see just the query without the ojects existing in the
 * SessionCache (as a result of creation).
 * 
 * @author martin
 *
 */
public class TopLinkTester {

	EntityManagerFactory factory = Persistence
			.createEntityManagerFactory("pu1");

	public static void main(String[] args) {
		
		boolean createData = false;
		if (args != null && args.length > 0) {
			createData = true;
		}
		TopLinkTester tester = new TopLinkTester();
		if (createData) {
			tester.createUsers();
			User u1 = tester.findUser("user1");
			User u2 = tester.findUser("user2");
			User u3 = tester.findUser("user3");
			tester.createMailMessages(u1, u2, u3);
		}
		
		tester.getMailMessages();
		
		// Uncomment to try again to see that info is not being
		// cached in the SessionCache.
		// Do the same thing again so can see any results of caching
		// tester.getMailMessages();

	}

	private void createUsers() {
		EntityManager manager = factory.createEntityManager();
		EntityTransaction tx = manager.getTransaction();
		tx.begin();
		try {

			User u1 = new User();
			u1.setLoginName("user1");

			User u2 = new User();
			u2.setLoginName("user2");

			User u3 = new User();
			u3.setLoginName("user3");

			manager.persist(u1);
			manager.persist(u2);
			manager.persist(u3);

			tx.commit();
			tx = null;
		} finally {
			if (tx != null) {
				try {
					tx.rollback();
				} catch (Exception e) {
					// ignore
				}
				manager.close();
			}
		}
	}

	private User findUser(String loginName) {
		User u = null;
		EntityManager manager = factory.createEntityManager();
		EntityTransaction tx = manager.getTransaction();
		tx.begin();
		try {
			u = findUserQuery(manager, loginName);

			tx.commit();
			tx = null;
		} finally {
			if (tx != null) {
				try {
					tx.rollback();
				} catch (Exception e) {
					// ignore
				}
				manager.close();
			}
		}
		return u;
	}

	private User findUserQuery(EntityManager manager, String loginName) {
		User u;
		Query query = manager.createQuery("SELECT u " + "FROM User u "
				+ "WHERE u.loginName = :name ");
		query.setParameter("name", loginName);
		u = (User) query.getSingleResult();
		return u;
	}

	private void createMailMessages(User u1, User u2, User u3) {
		EntityManager manager = factory.createEntityManager();
		EntityTransaction tx = manager.getTransaction();
		tx.begin();
		try {
			MailMessage message = new MailMessage();
			message.setSender(u1);
			message.setSubject("message1");
			manager.persist(message);

			MailMessage message2 = new MailMessage();
			message2.setSender(u1);
			message2.setSubject("message2");
			manager.persist(message2);

			tx.commit();
			tx = null;
		} finally {
			if (tx != null) {
				try {
					tx.rollback();
				} catch (Exception e) {
					// ignore
				}
				manager.close();
			}
		}

	}
	
	private void getMailMessages() {

		EntityManager manager = factory.createEntityManager();
		EntityTransaction tx = manager.getTransaction();
		tx.begin();
		try {
			
			// Update user 2 within the transaction to force
			// Toplink to start a real transaction when it
			// executes the queriers below. It is this which
			// prevents TopLink from writing the objects to the
			// SessionCache
			User u2 = findUserQuery(manager, "user2");
			u2.setFirstName(u2.getFirstName() + "1");
			
			User u1 = findUserQuery(manager, "user1");
			List<MailMessage> messages = getSent(manager, u1);
			for (MailMessage message : messages) {
				System.out.println(message.getSubject());
			}
			tx.commit();
			tx = null;
		} finally {
			if (tx != null) {
				try {
					tx.rollback();
				} catch (Exception e) {
					// ignore
				}
				manager.close();
			}
		}
	}

	@SuppressWarnings("unchecked")
	public List<MailMessage> getSent(EntityManager manager, User user) {
		Query query = manager
				.createQuery("SELECT m " + "FROM MailMessage m "
						+ "WHERE m.sender = :user ");
		query.setParameter("user", user);
		return query.getResultList();
	}
}
